package server

import (
	"bytes"
	"fmt"

	"github.com/labstack/echo/v4"
	"gitlab.bad-wolf.fr/wolf/hmac"
)

const (
	KeySignatureOK = "hmac-signature-ok"
)

// KeySignatureGetter return the key and signature associated to a given request
type KeySignatureGetter func(c echo.Context) (key, signature []byte, err error)

// SignatureChecker read the signature from a request and ensure it is legit
// This should be the first step on an authentication root
// If no error is returned by the KeySignatureGetter and key/signature are nil,
// the middleware will let the request in unless signatureRequired is true in the
func SignatureChecker(signer hmac.Signer, ksg KeySignatureGetter, signatureRequired bool) echo.MiddlewareFunc {
	return func(fn echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {

			key, receivedSignature, err := ksg(c)
			if err != nil {
				return fmt.Errorf(
					"unable to retrieve key and signature for request: %w", err)
			}

			if len(receivedSignature) == 0 {
				if signatureRequired {
					return echo.ErrUnauthorized
				}
				c.Set(KeySignatureOK, false)
				return fn(c)
			}

			properSignature, err := signer.Sign(c.Request(), key)
			if err != nil {
				return fmt.Errorf("unable to generate signature: %w", err)
			}

			if !bytes.Equal(properSignature, receivedSignature) {
				return echo.ErrUnauthorized
			}

			c.Set(KeySignatureOK, true)

			// All good, proceed !
			return fn(c)
		}
	}
}

func SignatureOK(c echo.Context) bool {
	res := c.Get(KeySignatureOK)
	switch res := res.(type) {
	case bool:
		return res
	case nil:
		return false
	default:
		panic(fmt.Errorf("%s does not contain a boolean but %T", KeySignatureOK, res))
	}
}

// SignatureEnforce is to be used after SignatureChecker to enforce a signature
// on a route, even though SignatureChecker signatureRequired params was false
func SignatureEnforce() echo.MiddlewareFunc {
	return func(fn echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			if !SignatureOK(c) {
				return echo.ErrUnauthorized
			}
			return fn(c)
		}
	}
}
