package requestencoder

import (
	"bytes"
	"net/http"
	"testing"

	"github.com/go-test/deep"
)

func TestTemplateEncoder(t *testing.T) {

	body := bytes.NewBufferString("HELLO WORLD IN BODY")
	req, err := http.NewRequest(http.MethodPost, "example.com/hello/world?test=ok&test2=okToo", body)
	if err != nil {
		t.Fatalf("unable to create the request: %s", err)
	}

	req.Header.Add("x-test-header-1", "test-value-1")
	req.Header.Add("x-test-header-2", "test-value-2")

	userData := map[string]string{
		"ignoredHeader":                "toto",
		"testURLWithQueryString":       "example.com/hello/world?test=ok&test2=okToo",
		"testURLWithNoQueryString":     "example.com/hello/world",
		"testURLWithQueryStringOnly":   "?test=ok&test2=okToo",
		"testURLWithQueryStringNoHost": "hello/world?test=ok&test2=okToo",
	}

	type args struct {
		tmplt string
		req   *http.Request
	}
	tests := []struct {
		name    string
		args    args
		want    []byte
		wantErr bool
	}{
		{
			"Test OK empty",
			args{
				"",
				req,
			},
			nil,
			false,
		},
		{
			"Test OK static string",
			args{
				"TEST",
				req,
			},
			[]byte(`TEST`),
			false,
		},
		{
			"Test OK request URL",
			args{
				"{{.Req.Method}};{{.Req.URL}}",
				req,
			},
			[]byte(`POST;example.com/hello/world?test=ok&test2=okToo`),
			false,
		},
		{
			"Test OK request headers",
			args{
				`{{range $key, $value := .Req.Header -}}
					{{- range $value -}}
						{{- $key -}}={{- . -}}
					{{- end -}}
				{{- end -}}`,
				req,
			},
			[]byte(`X-Test-Header-1=test-value-1X-Test-Header-2=test-value-2`),
			false,
		},
		{
			"Test OK request headers with separator",
			args{
				`{{range $key, $value := .Req.Header -}}
					{{- range $value -}}
						{{- $key -}}={{- . -}};
					{{- end -}}
				{{- end -}}`,
				req,
			},
			[]byte(`X-Test-Header-1=test-value-1;X-Test-Header-2=test-value-2;`),
			false,
		},
		{
			"Test OK with b64 encoding",
			args{
				`{{b64 "hello world"}}`,
				req,
			},
			[]byte(`aGVsbG8gd29ybGQ=`),
			false,
		},
		{
			"Test OK with md5 hash",
			args{
				`{{md5 "hello world" | b64}}`,
				req,
			},
			[]byte(`XrY7u+Ae7tCTyyK7j1rNww==`),
			false,
		},
		{
			"Test OK with sha256 hash",
			args{
				`{{sha256 "hello world" | b64}}`,
				req,
			},
			[]byte(`uU0nuZNNPgilLlLX2n2r+sSE7+N6U4DukIj3rOLvzek=`),
			false,
		},
		{
			"Test OK with sha512 hash",
			args{
				`{{sha512 "hello world" | b64}}`,
				req,
			},
			[]byte(`MJ7MSJwS1utMxA9QyQLytNDtd+5RGnx6m808qG1M2G+YndNbxf9JlnDaNCVbRbDP2DDoH2Bdz33FVC6TrpzXbw==`),
			false,
		},
		{
			"Test OK with user data",
			args{
				`{{.UserData.ignoredHeader}}`,
				req,
			},
			[]byte(`toto`),
			false,
		},
		{
			"Test OK regex testURLWithQueryString",
			args{
				`{{reReplace .UserData.testURLWithQueryString "(?:^[^?]+)?(?:\\?(.*))?$" "$1"}}`,
				req,
			},
			[]byte(`test=ok&test2=okToo`),
			false,
		},
		{
			"Test OK regex testURLWithNoQueryString",
			args{
				`{{reReplace .UserData.testURLWithNoQueryString "(?:^[^?]+)?(?:\\?(.*))?$" "$1"}}`,
				req,
			},
			[]byte(``),
			false,
		},
		{
			"Test OK regex testURLWithQueryStringOnly",
			args{
				`{{reReplace .UserData.testURLWithQueryStringOnly "(?:^[^?]+)?(?:\\?(.*))?$" "$1"}}`,
				req,
			},
			[]byte(`test=ok&test2=okToo`),
			false,
		},
		{
			"Test OK extract query string testURLWithQueryString",
			args{
				`{{extractQueryString .UserData.testURLWithQueryString}}`,
				req,
			},
			[]byte(`test=ok&test2=okToo`),
			false,
		},
		{
			"Test OK extract query string testURLWithNoQueryString",
			args{
				`{{extractQueryString .UserData.testURLWithNoQueryString}}`,
				req,
			},
			[]byte(``),
			false,
		},
		{
			"Test OK extract query string testURLWithQueryStringOnly",
			args{
				`{{extractQueryString .UserData.testURLWithQueryStringOnly}}`,
				req,
			},
			[]byte(`test=ok&test2=okToo`),
			false,
		},
		{
			"Test OK extract query string testURLWithQueryStringNoHost",
			args{
				`{{extractQueryString .UserData.testURLWithQueryStringNoHost}}`,
				req,
			},
			[]byte(`test=ok&test2=okToo`),
			false,
		},

		{
			"Test OK extract path testURLWithQueryString",
			args{
				`{{extractPath .UserData.testURLWithQueryString}}`,
				req,
			},
			[]byte(`example.com/hello/world`),
			false,
		},
		{
			"Test OK extract path testURLWithNoQueryString",
			args{
				`{{extractPath .UserData.testURLWithNoQueryString}}`,
				req,
			},
			[]byte(`example.com/hello/world`),
			false,
		},
		{
			"Test OK extract path testURLWithQueryStringOnly",
			args{
				`{{extractPath .UserData.testURLWithQueryStringOnly}}`,
				req,
			},
			[]byte(``),
			false,
		},
		{
			"Test OK extract path testURLWithQueryStringNoHost",
			args{
				`{{extractPath .UserData.testURLWithQueryStringNoHost}}`,
				req,
			},
			[]byte(`hello/world`),
			false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			e, err := TemplateEncoder(tt.args.tmplt, userData)

			if err != nil {
				t.Errorf("unable to parse template: %s", err)
				return
			}

			got, err := e(tt.args.req)

			if (err != nil) != tt.wantErr {
				t.Errorf("TemplateEncoder() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			// if !reflect.DeepEqual(got, tt.want) {

			if diffs := deep.Equal(string(got), string(tt.want)); diffs != nil {
				t.Errorf("TemplateEncoder() = %v, want %v (%s)", string(got), string(tt.want), diffs)
			}
		})
	}
}
